import math
from typing import List

from super_scad.d2.Polygon import Polygon
from super_scad.scad.ArgumentAdmission import ArgumentAdmission
from super_scad.scad.Context import Context
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.scad.Unit import Unit
from super_scad.type.Point2 import Point2


class RegularPolygon(ScadWidget):
    """
    Widget for creating regular polygons.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 sides: int,
                 outer_radius: float | None = None,
                 outer_diameter: float | None = None,
                 inner_radius: float | None = None,
                 inner_diameter: float | None = None,
                 size: float | None = None):
        """
        Object constructor.

        :param sides: The number of sider of the regular polygon.
        :param outer_radius: The outer radius (a.k.a. circumradius) of the regular polygon.
        :param inner_radius: The inner radius (a.k.a. apothem) of the regular polygon.
        :param size: The length of a side of the regular polygon.
        """
        ScadWidget.__init__(self, args=locals())

        self.__angles: List[float] = []
        """
        The angles of the nodes regular polygon.
        """

        self.__nodes: List[Point2] = []
        """
        The nodes of the regular polygon.
        """

        self.__unit: Unit | None = None
        """
        The unit in which self.__points are calculated.
        """

    # ------------------------------------------------------------------------------------------------------------------
    def _validate_arguments(self) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.
        """
        admission = ArgumentAdmission(self._args)
        admission.validate_exclusive({'size'},
                                     {'outer_radius'},
                                     {'outer_diameter'},
                                     {'inner_radius'},
                                     {'inner_diameter'})
        admission.validate_required({'sides'},
                                    {'size', 'outer_radius', 'outer_diameter', 'inner_radius', 'inner_diameter'})

    # ------------------------------------------------------------------------------------------------------------------
    def __nodes_and_angles(self) -> None:
        """
        Computes the nodes ond the angles of the nodes of the regular polygon.
        """
        if len(self.__angles) > 0 and self.__unit == Context.get_unit_length_current():
            return

        self.__unit = Context.get_unit_length_current()
        self.__nodes.clear()
        self.__angles.clear()

        step = 2.0 * math.pi / self.sides
        radius = self.outer_radius
        if self.sides % 2 == 0:
            # Even number of sides.
            angle = step / 2.0
        else:
            # Odd number of sides.
            angle = math.pi / 2.0

        for i in range(self.sides):
            self.__nodes.append(Point2(x=radius * math.cos(angle), y=radius * math.sin(angle)))
            self.__angles.append(math.degrees(angle))
            angle += step

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def sides(self) -> int:
        """
        Returns the number of sides of the polygon.
        """
        return self._args['sides']

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def outer_radius(self) -> float:
        """
        Returns the outer radius of the regular polygon.
        """
        if 'outer_radius' in self._args:
            return self.uc(self._args['outer_radius'])

        if 'outer_diameter' in self._args:
            return self.uc(0.5 * self._args['outer_diameter'])

        if 'inner_radius' in self._args:
            return self.uc(self._args['inner_radius'] / math.cos(math.pi / self.sides))

        if 'inner_diameter' in self._args:
            return self.uc(0.5 * self._args['inner_diameter'] / math.cos(math.pi / self.sides))

        if 'size' in self._args:
            return self.uc(self._args['size'] / (2.0 * math.sin(math.pi / self.sides)))

        raise AssertionError('Should not be reached')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def outer_diameter(self) -> float:
        """
        Returns the outer diameter of the regular polygon.
        """
        if 'outer_radius' in self._args:
            return self.uc(2.0 * self._args['outer_radius'])

        if 'outer_diameter' in self._args:
            return self.uc(self._args['outer_diameter'])

        if 'inner_radius' in self._args:
            return self.uc(2.0 * self._args['inner_radius'] / math.cos(math.pi / self.sides))

        if 'inner_diameter' in self._args:
            return self.uc(self._args['inner_diameter'] / math.cos(math.pi / self.sides))

        if 'size' in self._args:
            return self.uc(self._args['size'] / (math.sin(math.pi / self.sides)))

        raise AssertionError('Should not be reached')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def inner_radius(self) -> float:
        """
        Returns the inner radius of the regular polygon.
        """
        if 'inner_radius' in self._args:
            return self.uc(self._args['inner_radius'])

        if 'inner_diameter' in self._args:
            return self.uc(self._args['inner_diameter'] / 2.0)

        if 'outer_radius' in self._args:
            return self.uc(self._args['outer_radius'] * math.cos(math.pi / self.sides))

        if 'outer_diameter' in self._args:
            return self.uc(0.5 * self._args['outer_diameter'] * math.cos(math.pi / self.sides))

        if 'size' in self._args:
            return self.uc(self._args['size'] / (2.0 * math.tan(math.pi / self.sides)))

        raise AssertionError('Should not be reached')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def inner_diameter(self) -> float:
        """
        Returns the inner diameter of the regular polygon.
        """
        if 'inner_radius' in self._args:
            return self.uc(2.0 * self._args['inner_radius'])

        if 'inner_diameter' in self._args:
            return self.uc(self._args['inner_diameter'])

        if 'outer_radius' in self._args:
            return self.uc(2.0 * self._args['outer_radius'] * math.cos(math.pi / self.sides))

        if 'outer_diameter' in self._args:
            return self.uc(self._args['outer_diameter'] * math.cos(math.pi / self.sides))

        if 'size' in self._args:
            return self.uc(self._args['size'] / (math.tan(math.pi / self.sides)))

        raise AssertionError('Should not be reached')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def size(self) -> float:
        """
        Returns the inner radius of the regular polygon.
        """
        if 'inner_radius' in self._args:
            return self.uc(2.0 * self._args['inner_radius'] * math.tan(math.pi / self.sides))

        if 'inner_diameter' in self._args:
            return self.uc(self._args['inner_diameter'] * math.tan(math.pi / self.sides))

        if 'outer_radius' in self._args:
            return self.uc(2.0 * self._args['outer_radius'] * math.sin(math.pi / self.sides))

        if 'outer_diameter' in self._args:
            return self.uc(self._args['outer_diameter'] * math.sin(math.pi / self.sides))

        if 'size' in self._args:
            return self.uc(self._args['size'])

        raise AssertionError('Should not be reached')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def inner_angle(self) -> float:
        """
        Returns the inner angle in degrees between the edges of the regular polygon.
        """
        return 180.0 - 360.0 / self.sides

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def exterior_angle(self) -> float:
        """
        Returns the exterior angle in degrees between the edges of the regular polygon.
        """
        return 360.0 / self.sides

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def angles(self) -> List[float]:
        """
        Returns the angles in degrees of the position of the nodes of the regular polygon in polar coordinates.
        """
        self.__nodes_and_angles()

        return self.__angles

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def nodes(self) -> List[Point2]:
        """
        Returns the coordinates of the nodes of the regular polygon.
        """
        self.__nodes_and_angles()

        return self.__nodes

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        return Polygon(points=self.nodes)

# ----------------------------------------------------------------------------------------------------------------------



from typing import List

from super_scad.private.PrivateOpenScadCommand import PrivateOpenScadCommand
from super_scad.type.Face3 import Face3
from super_scad.type.Point3 import Point3


class Polyhedron(PrivateOpenScadCommand):
    """
    Widget for creating polyhedrons. See https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Primitive_Solids#polyhedron.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 points: List[Point3],
                 faces: List[Face3],
                 convexity: int | None = None):
        """
        Object constructor.

        :param points: Vector of 3d points or vertices. Each point is in turn a vector, [x,y,z], of its coordinates.
                       Points may be defined in any order. N points are referenced, in the order defined, as 0 to N-1.
        :param faces:  Vector of faces that collectively enclose the solid. Each face is a vector containing the
                       indices (0 based) of 3 or more points from the points vector. Faces may be defined in any order,
                       but the points of each face must be ordered correctly (see below). Define enough faces to fully
                       enclose the solid, with no overlap. If points that describe a single face are not on the same
                       plane, the face is automatically split into triangles as needed.
        :param convexity: Number of "inward" curves, i.e. expected number of path crossings of an arbitrary line through
                          the child widget.
        """
        PrivateOpenScadCommand.__init__(self, command='polyhedron', args=locals())

    # ------------------------------------------------------------------------------------------------------------------
    def _validate_arguments(self) -> None:
        pass

# ----------------------------------------------------------------------------------------------------------------------



import os
from pathlib import Path

from super_scad.private.PrivateMultiChildOpenScadCommand import PrivateMultiChildOpenScadCommand
from super_scad.private.PrivateOpenScadCommand import PrivateOpenScadCommand
from super_scad.private.PrivateSingleChildOpenScadCommand import PrivateSingleChildOpenScadCommand
from super_scad.scad.Context import Context
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.scad.Unit import Unit


class Scad:
    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self, unit_length_final: Unit):
        """
        Object constructor.

        :param unit_length_final: The unit of length used in the generated OpenSCAD code.
        """

        self.__project_home: Path = Path(os.getcwd()).resolve()
        """
        The current project's home directory.
        """

        self.__context = Context(project_home=self.__project_home, unit_length_final=unit_length_final)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def project_home(self) -> Path:
        """
        Returns the current project's home directory.
        """
        return self.__project_home

    # ------------------------------------------------------------------------------------------------------------------
    @project_home.setter
    def project_home(self, project_home: Path) -> None:
        """
        Sets the current project's home directory.

        :param Path project_home: The current project's home directory.
        """
        self.__project_home = project_home

    # ------------------------------------------------------------------------------------------------------------------
    def run_super_scad(self, root_widget: ScadWidget, openscad_path: Path | str) -> None:
        """
        Runs SuperSCAD on a SuperSCAD widget and stores the generated OpenSCAD code.

        :param root_widget: The root SuperSCAD widget to build.
        :param openscad_path: The path to the file were to store the generated OpenSCAD code.
        """
        self.__run_super_scad_prepare(openscad_path)
        self.__run_super_scad_walk_build_tree(root_widget)
        self.__run_super_scad_finalize()

    # ------------------------------------------------------------------------------------------------------------------
    def __run_super_scad_prepare(self, openscad_path: Path | str) -> None:
        """
        Executes the required steps before running SuperSCAD.

        :param openscad_path: The path to the file were to store the generated OpenSCAD code.
        """
        self.__context.target_path = Path(openscad_path)
        self.__context.set_unit_length_current(self.__context.get_unit_length_final())
        self.__context.code_store.clear()
        self.__context.code_store.add_line('// Unit of length: {}'.format(Context.get_unit_length_final()))

    # ------------------------------------------------------------------------------------------------------------------
    def __run_super_scad_finalize(self) -> None:
        """
        Executes the required step after running SuperSCAD.
        """
        self.__context.code_store.add_line('')

        with open(self.__context.target_path, 'wt') as handle:
            handle.write(self.__context.code_store.get_code())

    # ------------------------------------------------------------------------------------------------------------------
    def __run_super_scad_walk_build_tree(self, parent_widget: ScadWidget) -> None:
        """
        Helper method for __run_super_scad. Runs recursively on the SubSCAD widget and its children until it finds a
        widget for an OpenSCAD command. This OpenSCAD command is used to generate the OpenSCAD code.

        :param parent_widget: The parent widget to build.
        """
        old_unit = Context.get_unit_length_current()
        child_widget = parent_widget.build(self.__context)
        Context.set_unit_length_current(old_unit)

        if isinstance(child_widget, PrivateOpenScadCommand):
            self.__context.code_store.add_line('{}{}'.format(child_widget.command,
                                                             child_widget.generate_args(self.__context)))

            if isinstance(child_widget, PrivateSingleChildOpenScadCommand):
                self.__context.code_store.add_line('{')
                self.__run_super_scad_walk_build_tree(child_widget.child)
                self.__context.code_store.add_line('}')

            elif isinstance(child_widget, PrivateMultiChildOpenScadCommand):
                self.__context.code_store.add_line('{')
                for child in child_widget.children:
                    self.__run_super_scad_walk_build_tree(child)
                self.__context.code_store.add_line('}')

            else:
                self.__context.code_store.append_to_last_line(';')

        else:
            if child_widget == parent_widget:
                # Only OpenSCAD commands are allowed to build themselves.
                ValueError(f'Widget {parent_widget.__class__} build itself.')

            self.__run_super_scad_walk_build_tree(child_widget)

# ----------------------------------------------------------------------------------------------------------------------
